local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local L = LibStub("AceLocale-3.0"):GetLocale("RW2-Base")
local LB = LibStub("AceLocale-3.0"):GetLocale("RW2-Plugin-Base")

local AL = LibStub("AceLocale-3.0")
local LSM = LibStub("LibSharedMedia-3.0")
local LGT = LibStub("LibGroupTalents-1.0")

local Boss = {}
RW.BossPrototype = Boss
RW.Bosses:SetDefaultModulePrototype(Boss)

-- Party module, have all the info about the current party
local Party
local Comm
local Utils = RW.Utils

function Boss:OnInitialize()
	self.defaults = {
		enabled = true, 
		SortBossCategories = false,
		AutoHealthFrames = true,
	}
	
	self.BaseAlertObjects = {}
	self.AlertObjects = {}
	self.SyncSpamm = {}
	self.trackedSpells = {}
	
	-- Some constants
	self.pName = UnitName("player")
	self.pGUID = UnitGUID("player")
	self.pClass = select(2, UnitClass("player"))
	
	-- Added by kagaro: Use LibGroupTalents to determine role and monitor for respecs
	self.isCaster = LGT:GetUnitRole("player") == "caster"
	self.isMelee = LGT:GetUnitRole("player") == "melee"
	self.isTank = LGT:GetUnitRole("player") == "tank"
	self.isHealer = LGT:GetUnitRole("player") == "healer"
	LGT.RegisterCallback(self, "LibGroupTalents_RoleChange", "RoleChanged")

	RW:RegisterBoss(self) -- Calls OnRegister for the boss module and sets db
	
	self.db.RegisterCallback(self, "OnProfileChanged", "Reset")
	self.db.RegisterCallback(self, "OnProfileCopied", "Reset")
	self.db.RegisterCallback(self, "OnProfileReset", "Reset")
	self.DB = self.db.profile
	Party = RW.Party
	Comm = RW.Com
	
	self.Party = Party
	self.L = {}
end

function Boss:OnEnable()
	RW.RegisterCallback(self, "BarFinished") -- Handler declared in Timers prototype
	RW.Com:RegisterListener(self, "BossSync")
end

function Boss:OnDisable()
	RW.Com:UnregisterListener(self, "BossSync")	

	-- Cleanup
	self.engageTime = nil
	if self.trackedSpells then wipe(self.trackedSpells) end
	if self.trackedYells then wipe(self.trackedYells) end
	if self.trackedDeaths then wipe(self.trackedDeaths) end
	if self.SyncSpamm then wipe(self.SyncSpamm) end
	
	self.currentBossHP = nil
	
	-- Stop all alert objects and hide them
	self:UnscheduleAll()
	self:StopAllTimers()
	self:ResetAllIcons()
	self:HideAllDirections()
	RW.Callbacks:Fire("MPBHideAll")
	RW.Callbacks:Fire("RangeCheckFilterClear")
	RW.Callbacks:Fire("RangeCheckHide")	
end

function Boss:RoleChanged(guid, unit, role, oldrole)
	if guid == self.pGUID or unit == "player" then
		self.isCaster = role == "caster"
		self.isHealer = role == "healer"
		self.isTank = role == "tank"
		self.isMelee = role == "melee"
	end
end

function Boss:Reset()
	self.DB = self.db.profile
end

local function BossHealthScan(self)
	local hp = self:GetBossHealth()
	if hp ~= 0.0 then self.currentBossHP = hp end
end

function Boss:Start(trigger)
	if not self.TrashMod then
		if trigger then
			local bossName = RW.LBB[self:GetName()] or self:GetName()
			RW:Msg(L.BOSS_ENGAGE:format(bossName))
			RW.Callbacks:Fire("WarningShow", L.BOSS_ENGAGE:format(bossName), self.BossIcon, nil, "None")
		end
	end
	self.engageTime = GetTime()
	
	self:ScheduleRepeating("BossHealthScan", 3, BossHealthScan, nil, self)
	
	if RW.db.profile.GlobalBossHealthFrames and self.db.profile.AutoHealthFrames and not self.SupressGlobalHealthFrames then	
		for mob in pairs(self.mobTriggers) do
			RW.Callbacks:Fire("MPBHealthByID", mob, "Health Frames")
		end
	end
	-- we determined we are starting a boss, make sure the combatlog is clear
	CombatLogClearEntries()	
	self:Enable()
	if self.OnEngage then self:OnEngage(trigger) end
end

function Boss:End(wipe, trigger)
	if not self.TrashMod then
		local timeTaken
		if self.engageTime then
			timeTaken = GetTime() - self.engageTime
		else
			timeTaken = 0
		end

		if trigger then
			local bossName = RW.LBB[self:GetName()] or self:GetName()
			-- Did not wipe
			if not wipe then
				RW:Msg(L.BOSS_WIN:format(bossName, Utils:TimeToStr(timeTaken)))
				PlaySoundFile(LSM:Fetch("sound", RW.db.profile.vicotorySound))
				RW.Callbacks:Fire("WarningShow", (bossName.." defeated!"), self.BossIcon, nil, "None")				
			-- Wipe
			else
				RW:Msg(L.BOSS_WIPE:format(bossName, Utils:TimeToStr(timeTaken), (self.currentBossHP or 1) * 100))
			end
		end
	end
	if self.OnEnd then 
		self:OnEnd(wipe, trigger) 
	end
	self:Disable()
end

function Boss:RegisterLocale(lang)
	--@non-debug@
	return AL:NewLocale("RW-Boss-"..self:GetName(), lang, true, true)
	--@end-non-debug@
	--[===[@debug@
	return AL:NewLocale("RW-Boss-"..self:GetName(), lang, true)
	--@end-debug@]===]
end

function Boss:GetLocale()
	--@non-debug@
	return AL:GetLocale("RW-Boss-"..self:GetName(), true)
	--@end-non-debug@
	--[===[@debug@
	return AL:GetLocale("RW-Boss-"..self:GetName())
	--@end-debug@]===]
end

function Boss:SetAsTrashMod()
	self.TrashMod = true
end

--[[ function Boss:GetOpt(optName)
	return self.db.profile[optName]
end ]]--

function Boss:SetOrder(order)
	self.order = order
end

function Boss:SetDelay(delay)
	self.endDelay = delay
end

function Boss:ToggleHold()
	if self.hold then 
		self.hold = nil
	else
		self.hold = true
	end
end

function Boss:SetPackIcon(icon)
	self.PackIcon = icon
end

--[[
	New trigger functions
--]]

function Boss:AddEmoteTrigger(...)
	for i = 1, select('#',...) do
		local v = select(i,...)
		assert(type(v) == "string", self:GetName()..": EmoteTrigger is not of type text")
		self.yellTriggers = self.yellTriggers or {}
		self.yellTriggers[v] = true
	end
end

function Boss:AddCombatTrigger(...)
	local event = select(1,...)
	assert(type(event) == "string",self:GetName()..": Usage self:AddCombatTrigger(event_name,boss_id_number,spell_ids)")
	local mob = select(2,...)
	assert(type(mob) == "number" or mob == nil ,self:GetName()..": Usage self:AddCombatTrigger(event_name,boss_id_number,spell_ids|spell_names)")
	local spellIds = {}
	local idx = 1
	for i=3, select('#',...) do
		spellIds[idx] = select(i,...)
		idx = idx+1
	end
	if #spellIds == 0 and select('#',...) == 3 then
		spellIds[1] = select(3,...)
	end
	assert(#spellIds > 0,self:GetName()..": Usage self:AddCombatTrigger(event_name,boss_id_number,spell_ids)")
	self.combatTriggers = self.combatTriggers or {}
	self.combatTriggers[event] = {
		uid = mob,
		spells = spellIds
	}
	RW.Callbacks:Fire("CombatTrackingRegistered", self)	
end

function Boss:AddMobTrigger(...)
	for i = 1, select('#',...) do
		local v = select(i,...)
		assert(type(v) == "number", self:GetName()..": MobTrigger is not of type number")
		self.mobTriggers = self.mobTriggers or {}
		self.mobTriggers[v] = true
	end
end

function Boss:SetTrigger(...)
	for i = 1, select("#", ...) do
		local v = select(i, ...)
		if type(v) == "string" then
			self.yellTriggers = self.yellTriggers or {}
			self.yellTriggers[v] = true
		elseif type(v) == "number" then
			self.mobTriggers = self.mobTriggers or {}
			self.mobTriggers[v] = true
		end
	end
end

function Boss:SetWin(...)
	if not ... then
		self.mobWins = self.mobWins or {}
		local first = self:GetFirstTrigger()
		if first then
			self.mobWins[self:GetFirstTrigger()] = true
		end
	elseif select(1, ...) == "combat" then
		self.winTrigger = "combat"
	end
	for i = 1, select("#", ...) do
		local v = select(i, ...)
		if type(v) == "string" then
			self.yellWins = self.yellWins or {}
			self.yellWins[v] = true
		elseif type(v) == "number" then
			self.mobWins = self.mobWins or {}
			self.mobWins[v] = true
		end
	end
end

function Boss:GetFirstTrigger()
	if not self.mobTriggers or not type(self.mobTriggers) == "table" then return end
	local cID
	for mob in pairs(self.mobTriggers) do
		cID = mob
		break
	end
	return cID
end

-------------------------------------
-- Sync functions
function Boss:RemoveSyncSpamm(id)
	if not id then return end
	self.SyncSpamm[id] = nil
end

function Boss:SendSync(...)
	local id = ""
	local currentTime = time()
	for i=1, select("#", ...) do
		local var = select(i, ...)
		if type(var) == "string" or type(var) == "number" then
			id = id..var
		end
	end
	if not self.SyncSpamm[id] or currentTime - self.SyncSpamm[id] > 2 then
		self.SyncSpamm[id] = currentTime
		self:Schedule("RemoveSync"..id, 5, "RemoveSyncSpamm", id)
		self:BossSync(self.pName, "WHISPER", ...)
		RW.Com:Send("BossSync", nil, nil, ...)
	end
end

function Boss:SendWhisperSync(target, ...)
	RW.Com:Send("BossSync", "WHISPER", target, ...)
end

function Boss:BossSync(sender, dist, msg, ...)
	local id = msg or ""
	for i=1, select("#", ...) do
		local var = select(i, ...)
		if type(var) == "string" or type(var) == "number" then
			id = id..var
		end
	end
	if (not self.SyncSpamm[id] or (time() - self.SyncSpamm[id] > 5)) and self.OnSync then
		self:OnSync(sender, msg, ...)
	end
end

-------------------------------------
-- Party functions
function Boss:GetPartyType()
	return Party:GetPartyType()
end

function Boss:IsHeroic()
	return Utils:IsHeroic()
end

function Boss:InstanceSize()
	return Utils:InstanceSize()
end

function Boss:IsInstanceType(...)
	local name, instanceType, difficultyIndex, difficultyName, maxPlayers = GetInstanceInfo()
	local iString = (self:IsHeroic() and "heroic" or "normal")..maxPlayers
	for i = 1, select("#", ...) do
		if iString == slect(i, ...) then return true end
	end
	
	return false
end

-------------------------------------
-- cID functions
function Boss:CUIDFromGUID(gUID)
	if not gUID then return end
	-- CATA (GUID structure changed)
	if select(4, GetBuildInfo()) < 30400 then
		return tonumber(gUID:sub(9, 12), 16)
	else
		return tonumber(gUID:sub(7, 10), 16)
	end
end

function Boss:CUIDFromUID(uID)
	if not uID then return end
	local gUID = UnitGUID(uID)
	if gUID then
		return self:CUIDFromGUID(gUID)
	end
end

-------------------------------------
-- uID Functions
function Boss:UIDFromCID(cID)
	if self:CUIDFromGUID(UnitGUID("mouseover")) == cID then
		return "mouseover"
	end
	for name, member in pairs(Party.group) do
		if self:CUIDFromGUID(UnitGUID(member.id.."target")) == cID then
			return member.id.."target"
		end
	end
end

function Boss:UIDFromGUID(gUID)
	if not gUID then return end
	if UnitGUID("mouseover") == gUID then
		return "mouseover"
	end
	for name, member in pairs(Party.group) do
		if UnitGUID(member.id.."target") == gUID then
			return member.id.."target"
		end		
	end
end

function Boss:UIDFromName(name)
	if UnitName("mouseover") == name then return "mouseover" end
	if UnitName("player") == name then return "player" end
	for name, member in pairs(Party.group) do
		if UnitName(member.id) == name then return member.id end
		if UnitName(member.id.."target") == name then return member.id.."target" end
	end
end

function Boss:UIDFromRandom(target)
	local targetID
	if self:IsUID(target) then
		return target
	elseif target:find("^0x") then				-- Target by gUID
		targetID = self:UIDFromGUID(target)
	elseif Party.group[target] then				-- Target by party member name
		targetID = Party.group[target].id
	else		
		targetID = self:UIDFromName(target)  	-- Target by mob name
	end
	return targetID
end

function Boss:IsUID(arg)
	if type(arg) == "string" then
		if arg:find("^raid") or arg:find("^party") or arg:find("^target") or arg:find("^player") or arg:find("^boss") or arg:find("focus^") or arg:find("mouseover") then
			return true
		end
	end
	return false
end

-- Target
function Boss:GetMobTarget(cID)
	if not cID then cID = self:GetFirstTrigger() end
	if not cID then return end
	if cID == Boss:CUIDFromUID("target") then return UnitName("targettarget") end
	for i, player in pairs(Party.group) do
		if cID == Boss:CUIDFromUID(player.id.."target") then return UnitName(player.id.."targettarget") end
	end
end

-- Health
function Boss:GetBossHealth(cID)
	cID = cID or self:GetFirstTrigger()
	if not cID then return 0.0 end
	local uID = self:UIDFromCID(cID)
	if not uID then return 0.0 end
	return UnitHealth(uID) / UnitHealthMax(uID)
end

function Boss:SetIcon(icon, target)
	SetRaidTarget(target, icon or 0)
end

-------------------------------------
-- Tacking

function Boss:TrackAuraApplied(func,...)
    self:_TrackAura(func, "SPELL_AURA_APPLIED", "SPELL_AURA_APPLIED_DOSE", "SPELL_AURA_REFRESHED", ...)
end

function Boss:TrackAuraRemoved(func,...)
    self:_TrackAura(func, "SPELL_AURA_REMOVED", "SPELL_AURA_REMOVED_DOSE", "SPELL_AURA_BROKEN", ...)
end

function Boss:TrackAura(func, ...)
    self:_TrackAura(func, "SPELL_AURA_APPLIED", "SPELL_AURA_APPLIED_DOSE", "SPELL_AURA_REFRESHED", "SPELL_AURA_REMOVED", "SPELL_AURA_REMOVED_DOSE", "SPELL_AURA_BROKEN", ...)
end

function Boss:TrackCastStart(func, ...)
    self:_TrackCast(func, "SPELL_CAST_START", ...)
end

function Boss:TrackCastComplete(func,...)
    self:_TrackCast(func, "SPELL_CAST_SUCCESS", ...)
end	

function Boss:TrackCast(func, ...)
    self:_TrackCast(func, "SPELL_CAST_SUCCESS", "SPELL_CAST_START", ...)
end

--[[
	@args func to call
	@vargs any string args will be assumed to be the events to filter for, numebr for the spells
--]]
local function unpack2(t) return unpack(t, 1, t.n) end

function Boss:_TrackCast(func,...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	local events = {}
	local spellIdOffset = 1
	for i = 1, select("#",...) do
		local v = select(i,...)
		if type(v) == "string" then
			events[v] = true
		else
			spellIdOffset = i
			break
		end
	end
	local fun = function(boss,event,args)
		if not events[event] then return end
		boss[func](boss,args)
	end
	self.trackedSpells = self.trackedSpells or {}
	for i = spellIdOffset, select("#", ...) do
		self.trackedSpells[select(i, ...)] = fun
	end
	local n = select('#', ...)
	local m = 1
  	local t = {}
	for i = spellIdOffset,n do
    	t[m] = select(i,...)
		m = m + 1
	end
	RW.Callbacks:Fire("TrackSpell", self, fun, unpack2(t))	
end

function Boss:_TrackAura(func,...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	local events = {}
	local spellIdOffset = 1
	for i = 1, select("#",...) do
		local v = select(i,...)
		if type(v) == "string" then
			events[v] = true
		else
			spellIdOffset = i
			break
		end
	end
	local fun = function(boss,event,args)
		if not events[event] then return end
		if event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_APPLIED_DOSE" or event == "SPELL_AURA_REFRESHED" then
			boss[func](boss, true, args)
		else
			boss[func](boss, false,args)
		end
	end
	self.trackedSpells = self.trackedSpells or {}
	for i = spellIdOffset, select("#", ...) do
		self.trackedSpells[select(i, ...)] = fun
	end
	-- We do this so we dont see non spells here
	local n = select('#', ...)
	local m = 1
  	local t = {}
	for i = spellIdOffset,n do
    	t[m] = select(i,...)
		m = m + 1
	end
	RW.Callbacks:Fire("TrackSpell", self, fun, unpack2(t))	
end
function Boss:Track(func, ...)
	if type(func) == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	RW.Callbacks:Fire("TrackSpell", self, func, ...)
	self.trackedSpells = self.trackedSpells or {}
	for i = 1, select("#", ...) do
		self.trackedSpells[select(i, ...)] = func
	end
end

function Boss:YellTrack(func, ...)
	if type(func) == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	RW.Callbacks:Fire("TrackYell", self, func, ...)
	self.trackedYells = self.trackedYells or {}
	for i = 1, select("#", ...) do
		self.trackedYells[select(i, ...)] = func
	end
end

function Boss:DeathTrack(func, ...)
	if type(func) == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	RW.Callbacks:Fire("TrackDeath", self, func, ...)
	self.trackedDeaths = self.trackedDeaths or {}
	for i = 1, select("#", ...) do
		self.trackedDeaths[select(i, ...)] = func
	end
end


-------------------------------------
-- Scheduling

do
	local timers = {}
	local args = {}
		
	local function clearTimer(id)
		if not id or not timers[id] then return end
		local timer = timers[id]
		if timer.args then wipe(timer.args) end
		wipe(timer)
		timers[id] = nil
	end
	
	local function handleTimers(id)
		local timer = timers[id]
		local func = timer.func
		local m = timer.module
		if timer.args then
			wipe(args)
			for i, v in pairs(timer.args) do tinsert(args, v) end
		end
		if type(timer.func) == "function" then
			if not timer.rep then
				clearTimer(id)	
			end
			func(unpack(args))
		else
			if not timer.rep then
				clearTimer(id)	
			end
			if m[func] then
				m[func](m, unpack(args))
			end
		end
	end
	
	local function prepareSchedule(self, id, time, func, ...)
		local timer = timers[id]
		if timer then
			self:Unschedule(id)
		else
			timer = {}
		end
		timer.func = func
		timer.module = self
		timer.args = {}
		if select("#", ...) > 0 then
			if timer.args then
				for i = 1, select("#", ...) do
					tinsert(timer.args, (select(i, ...)))
				end
			else
				timer.args = {...}
			end
		end
		timers[id] = timer
	end
	
	function Boss:Schedule(id, time, func, ...)
		prepareSchedule(self, id, time, func, ...)
		timers[id].id = self:ScheduleTimer(handleTimers, time, id)
	end
	
	function Boss:ScheduleRepeating(id, time, func, now, ...)
		prepareSchedule(self, id, time, func, ...)
		timers[id].rep = true
		timers[id].id = self:ScheduleRepeatingTimer(handleTimers, time, id)
		if now then handleTimers(id) end
	end

	function Boss:Unschedule(id)
		if timers[id] then
			self:CancelTimer(timers[id].id, true)
			clearTimer(id)
		end
	end

	function Boss:UnscheduleAll()
		for id, timer in pairs(timers) do
			if timer.module == self then
				self:Unschedule(id)
			end
		end
	end
end

function Boss:AddBoolOption(varName, optionName, category, default, icon)
	local base = {
		type = "bool",
		icon = icon,			
		id = optionName,
		var = varName,
		mod = self,
	}
	self.defaults[varName] = ((default == nil) or default)
	tinsert(self.BaseAlertObjects, base)
	self:AddToCategory(category, optionName)
end

function Boss:AddRangeCheckOption(optName, reason, range, default, cat, icon)
	self:AddBoolOption(optName, LB.RANGE_CHECK_OPTION:format(reason, range), cat, default, icon or [[Interface\Icons\Achievement_BG_returnXflags_def_WSG]])
end

function Boss:AddDefaults(optionName, aType, sType, default)
	self.defaults[optionName] = self.defaults[optionName] or {}	
	self.defaults[optionName][aType] = self.defaults[optionName][aType] or {}
	self.defaults[optionName][aType][sType] = {
		enabled = ((default == nil) or default)
	}
	self.defaults[optionName].enabled = true
	
	return self.defaults[optionName][aType][sType]
end

do
	local function AddToGeneral(self, ...)
		local genCat
		if not Utils:InTable2(self.Options.Categories, "text", "General") then
			genCat = {text = "General"}
			tinsert(self.Options.Categories, genCat)
		else
			genCat = Utils:GetTableData(self.Options.Categories, "text", "General")
		end
		for i=1, select("#", ...) do
			local used = false
			local newopt = select(i, ...)
			for _, cat in pairs(self.Options.Categories) do
				for j, opt in pairs(cat) do
					if opt == newopt then used = true end
				end
			end
			if genCat then
				if not used then tinsert(genCat, newopt) end
			end
		end
	end

	function Boss:AddToCategory(cat, ...)
		self.Options = self.Options or {}
		self.Options.Categories = self.Options.Categories or {}
		
		if not cat then return AddToGeneral(self, ...) end
			
		local newCat
		if Utils:InTable2(self.Options.Categories, "text", cat) then
			for _, oldcat in pairs(self.Options.Categories) do				
				if oldcat.text == cat then
					newCat = oldcat
				end
			end
		else
			newCat = {
				text = cat
			}
			tinsert(self.Options.Categories, newCat)
		end		
		
		for i=1, select("#", ...) do
			local opt = select(i, ...)
			tinsert(newCat,  type(opt) == "number" and Utils.SpellName[opt] or opt)
		end
	end
end

function Boss:GetBaseAlertObject(id)
	for i, obj in pairs(self.BaseAlertObjects) do
		if obj.id == id then return obj end
	end
end

function Boss:GetIconObject(id)
	if not self.AlertObjects.ICON then return end
	for i, obj in pairs(self.AlertObjects.ICON) do
		if id == obj.base.id then return obj end
	end
end
